package cc.javastudio.raw;
import java.io.*;
import java.lang.reflect.Field;
import java.nio.ByteBuffer;
import java.util.HashMap;

public class FileHandler implements AutoCloseable {
    private RandomAccessFile dbFile;
    private String dbFileName;
    private final int INT_LENGTH = 4;
    private final int BYTE_LENGTH = 1;

    public FileHandler(String dbFileName) throws FileNotFoundException {
        this.dbFileName = dbFileName;
        dbFile = new RandomAccessFile(dbFileName, "rw");
    }

    /*
    Writes data as follows:
    name = "John Doe"
    age = 42
    address = "Berlin"
    licensePlate = "BMW-123"
    description = "Fiat Punto"
    sizes = 8;4;6;7;10 = 35
    output = 0358John Doe426Berlin7BMW-12310Fiat Punto
     */
    public void add(String name,
                    int age,
                    String address,
                    String licensePlate,
                    String description) throws IOException {
        // Set the pointer at the end of the current file
        dbFile.seek(dbFile.length());

        // Create block starters (metadata)
        /*
        // This block can't handle UTF8 (2 bytes for a special char - eg ñ - & 1 byte for normal characters)
        int lenName = name.length(),
            lenAddress = address.length(),
            lenLicensePlate = licensePlate.length(),
            lenDescription = description.length();
         */
        int lenName = name.getBytes().length,
            lenAddress = address.getBytes().length,
            lenLicensePlate = licensePlate.getBytes().length,
            lenDescription = description.getBytes().length;

        // Length in bytes - 1 char = 1 byte, 1 integer = 4 bytes, ...
        int recordLength = INT_LENGTH + lenName +
                INT_LENGTH + // age length not necessary - we'll write the age number directly
                INT_LENGTH + lenAddress +
                INT_LENGTH + lenLicensePlate +
                INT_LENGTH + lenDescription;

        // Write data in a format [deleted flag][record length][data length][data]...[nth data length][nth data]
        dbFile.writeBoolean(false); // write deleted flag - false = not deleted
        dbFile.writeInt(recordLength);

        dbFile.writeInt(lenName);
        dbFile.write(name.getBytes());

        dbFile.writeInt(age);

        dbFile.writeInt(lenAddress);
        dbFile.write(address.getBytes());

        dbFile.writeInt(lenLicensePlate);
        dbFile.write(licensePlate.getBytes());

        dbFile.writeInt(lenDescription);
        dbFile.write(description.getBytes());
    }

    public Person readRow(int rowNumber) throws IOException {
        Person person = new Person();
        byte[] rawData = readRawRecord(rowNumber);
        DataInputStream stream = new DataInputStream(new ByteArrayInputStream(rawData));

        byte[] partial = new byte[stream.readInt()];
        stream.read(partial);
        person.name = new String(partial);

        partial = new byte[INT_LENGTH];
        stream.read(partial);
        person.age = ByteBuffer.wrap(partial).getInt();

        partial = new byte[stream.readInt()];
        stream.read(partial);
        person.address = new String(partial);

        partial = new byte[stream.readInt()];
        stream.read(partial);
        person.licensePlate = new String(partial);

        partial = new byte[stream.readInt()];
        stream.read(partial);
        person.description = new String(partial);

        return person;
    }

    public byte[] readRawRecord(int recordNumber) throws IOException {
        byte[] data;
        int idx = 0;
        int pos = 0;
        int lenRecord = -1;

        while(idx <=recordNumber) {
            dbFile.seek(pos);
            pos += BYTE_LENGTH;

            if (dbFile.readBoolean()) {
                //dbFile.seek(pos);   // optional: readBoolean sets pointer at pos + 1 byte
                lenRecord = dbFile.readInt();
                pos += lenRecord;
            }
            else {
                Index.getInstance().set(idx, pos);
                //dbFile.seek(pos);   // optional: readBoolean sets pointer at pos + 1 byte
                lenRecord = dbFile.readInt();
                pos += INT_LENGTH;
                //dbFile.seek(pos);   // optional: readInt sets pointer at pos + 4 bytes

                if (idx == recordNumber) {
                    data = new byte[lenRecord];
                    dbFile.read(data);
                    return data;
                }
                else {
                    pos += lenRecord;
                }

                idx++;
            }
        }

        return null;
    }

    public byte[] readAll() throws IOException {
        byte[] data = new byte[(int)dbFile.length()];
        dbFile.readFully(data);
        return data;
    }

    public void close() throws IOException {
        dbFile.close();
    }
}
